{
 "cells": [
  {
   "cell_type": "markdown",
   "id": "02c22b4d-d9d8-4d56-b1fb-298b1106f3b0",
   "metadata": {},
   "source": [
    "# Ray and Daft\n",
    "\n",
    "Example using Daft on a Ray cluster to query a Vast DB table and return it as a [Daft](https://www.getdaft.io/projects/docs/en/stable/api_docs/doc_gen/dataframe_methods/daft.DataFrame.html) Dataframe.\n",
    "\n",
    "NOTE: \n",
    "- This notebook needs to be run with the same version of Python as the Ray Cluster.\n",
    "- This notebook does not parallelize reading data from Vast DB."
   ]
  },
  {
   "cell_type": "markdown",
   "id": "8492d63b-1d71-4a61-8524-197455011bd0",
   "metadata": {},
   "source": [
    "## Install dependencies"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "id": "a00577bd-011d-42cc-ae95-22fef92791f5",
   "metadata": {},
   "outputs": [],
   "source": [
    "# pip uninstall -y \"daft[ray]\" \"ray[client]==2.41.0\""
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "id": "686c542f-a848-41d9-8cab-13338a726461",
   "metadata": {},
   "outputs": [],
   "source": [
    "%%capture --no-stderr\n",
    "%pip install \"ray[client,train]==2.41.0\"\n",
    "%pip install \"daft[ray]\" --no-deps\n",
    "%pip install vastdb"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "id": "3cb82aba-0350-4f6f-be19-d4ec06bc2ab4",
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "daft                      0.4.8\n",
      "ray                       2.41.0\n"
     ]
    }
   ],
   "source": [
    "!pip list | grep -E \"(ray|daft)\""
   ]
  },
  {
   "cell_type": "markdown",
   "id": "7c4fcbe1-a78d-4216-b3c2-08425bb9dace",
   "metadata": {},
   "source": [
    "## Create Vast DB Ray DataSource"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "id": "150f835d-6191-4d66-9836-68b84b638970",
   "metadata": {},
   "outputs": [],
   "source": [
    "import pyarrow as pa\n",
    "import vastdb\n",
    "import daft\n",
    "import ray\n",
    "import traceback\n",
    "import sys\n",
    "import time\n",
    "from typing import Dict, Any, Optional\n",
    "\n",
    "@ray.remote\n",
    "class VastDataSource:\n",
    "    def __init__(\n",
    "        self, \n",
    "        endpoint: str, \n",
    "        access_key: str, \n",
    "        secret_key: str,\n",
    "        bucket_name: str, \n",
    "        schema_name: str, \n",
    "        table_name: str\n",
    "    ):\n",
    "        \"\"\"\n",
    "        Initialize VAST data source connection to be run on Ray cluster\n",
    "        \"\"\"\n",
    "        self.endpoint = endpoint\n",
    "        self.access_key = access_key\n",
    "        self.secret_key = secret_key\n",
    "        self.bucket_name = bucket_name\n",
    "        self.schema_name = schema_name\n",
    "        self.table_name = table_name\n",
    "        \n",
    "        # Verbose print for initialization\n",
    "        print(f\"[INIT] VastDataSource initialized for table: {table_name}\", file=sys.stderr)\n",
    "    \n",
    "    def verbose_batch_reading(self, reader: pa.RecordBatchReader):\n",
    "        \"\"\"\n",
    "        Verbose batch reading with detailed diagnostics\n",
    "        \"\"\"\n",
    "        print(f\"[DEBUG] Starting batch reading for {self.table_name}\", file=sys.stderr)\n",
    "        \n",
    "        try:\n",
    "            import socket\n",
    "            import os\n",
    "        \n",
    "            # Capture host information\n",
    "            hostname = socket.gethostname()\n",
    "            ip_address = socket.gethostbyname(hostname)\n",
    "            pid = os.getpid()\n",
    "            \n",
    "            # Print detailed host information\n",
    "            print(f\"[HOST] Hostname: {hostname}\", file=sys.stderr)\n",
    "            print(f\"[DEBUG] Reader type: {type(reader)}\", file=sys.stderr)\n",
    "            print(f\"[DEBUG] Reader schema: {reader.schema}\", file=sys.stderr)\n",
    "            \n",
    "            # Batch collection with verbose logging\n",
    "            batches = []\n",
    "            batch_count = 0\n",
    "            start_time = time.time()\n",
    "            \n",
    "            try:\n",
    "                # Attempt different iteration methods\n",
    "                for batch in reader:\n",
    "                    print(f\"[DEBUG] Collected batch {batch_count}\", file=sys.stderr)\n",
    "                    batches.append(batch)\n",
    "                    batch_count += 1\n",
    "                    \n",
    "                    # Safety break to prevent infinite loops\n",
    "                    if batch_count >= 1000:  # Adjust as needed\n",
    "                        print(f\"[WARNING] Reached batch limit of 1000\", file=sys.stderr)\n",
    "                        break\n",
    "            except Exception as iter_err:\n",
    "                print(f\"[ERROR] Iteration error: {iter_err}\", file=sys.stderr)\n",
    "                traceback.print_exc(file=sys.stderr)\n",
    "            \n",
    "            # Timing and summary\n",
    "            end_time = time.time()\n",
    "            print(f\"[DEBUG] Batch reading summary:\", file=sys.stderr)\n",
    "            print(f\"  Total batches: {batch_count}\", file=sys.stderr)\n",
    "            print(f\"  Total time: {end_time - start_time:.2f} seconds\", file=sys.stderr)\n",
    "            \n",
    "            return batches\n",
    "        \n",
    "        except Exception as e:\n",
    "            print(f\"[CRITICAL] Batch reading failed: {e}\", file=sys.stderr)\n",
    "            traceback.print_exc(file=sys.stderr)\n",
    "            return []\n",
    "\n",
    "    def get_vast_data(self, columns=None, predicate=None, internal_row_id=False):\n",
    "        \"\"\"\n",
    "        Retrieve data from VAST database with verbose error handling\n",
    "        \"\"\"\n",
    "        print(f\"[METHOD] get_vast_data called for {self.table_name}\", file=sys.stderr)\n",
    "        print(f\"[PARAMS] Columns: {columns}, Predicate: {predicate}\", file=sys.stderr)\n",
    "\n",
    "        try:\n",
    "            # Connection establishment with verbose logging\n",
    "            try:\n",
    "                session = vastdb.connect(\n",
    "                    endpoint=self.endpoint,\n",
    "                    access=self.access_key,\n",
    "                    secret=self.secret_key\n",
    "                )\n",
    "                print(\"[CONNECTION] VAST database connection established\", file=sys.stderr)\n",
    "            except Exception as conn_err:\n",
    "                print(f\"[ERROR] Connection failed: {conn_err}\", file=sys.stderr)\n",
    "                traceback.print_exc(file=sys.stderr)\n",
    "                raise\n",
    "            \n",
    "            with session.transaction() as tx:\n",
    "                bucket = tx.bucket(self.bucket_name)\n",
    "                schema = bucket.schema(self.schema_name)\n",
    "                table = schema.table(self.table_name)\n",
    "                \n",
    "                try:\n",
    "                    # Select data with comprehensive error handling\n",
    "                    reader: pa.RecordBatchReader = table.select(\n",
    "                        columns=columns, \n",
    "                        predicate=predicate,\n",
    "                        internal_row_id=internal_row_id\n",
    "                    )\n",
    "                    \n",
    "                    # Verbose batch reading\n",
    "                    batches = self.verbose_batch_reading(reader)\n",
    "                    \n",
    "                    if not batches:\n",
    "                        print(\"[WARNING] No batches collected\", file=sys.stderr)\n",
    "                        return pa.Table.from_batches([])\n",
    "                    \n",
    "                    # Convert to table\n",
    "                    result_table = pa.Table.from_batches(batches)\n",
    "                    \n",
    "                    print(f\"[SUCCESS] Retrieved table with {result_table.num_rows} rows\", file=sys.stderr)\n",
    "                    return result_table\n",
    "                \n",
    "                except Exception as select_err:\n",
    "                    print(f\"[ERROR] Data selection failed: {select_err}\", file=sys.stderr)\n",
    "                    traceback.print_exc(file=sys.stderr)\n",
    "                    raise\n",
    "        \n",
    "        except Exception as e:\n",
    "            print(f\"[CRITICAL] Comprehensive error in get_vast_data: {e}\", file=sys.stderr)\n",
    "            traceback.print_exc(file=sys.stderr)\n",
    "            raise ValueError(f\"Error reading data from VAST: {e}\")\n",
    "\n",
    "def fetch_as_daft_dataframe(\n",
    "    endpoint: str, \n",
    "    access_key: str, \n",
    "    secret_key: str,\n",
    "    bucket_name: str, \n",
    "    schema_name: str, \n",
    "    table_name: str,\n",
    "    columns: Optional[list] = None,\n",
    "    predicate: Optional[Any] = None,\n",
    "    internal_row_id: Optional[bool] = None,\n",
    "    context = None\n",
    ") -> daft.DataFrame:\n",
    "    \"\"\"\n",
    "    Create a Daft DataFrame from a VAST data source running on Ray cluster\n",
    "    \"\"\"\n",
    "    print(f\"[FUNCTION] create_vast_table called for {table_name}\", file=sys.stderr)\n",
    "    \n",
    "    try:\n",
    "        # Create a Ray actor for the VAST data source\n",
    "        vast_actor = VastDataSource.remote(\n",
    "            endpoint, access_key, secret_key, \n",
    "            bucket_name, schema_name, table_name\n",
    "        )\n",
    "        \n",
    "        # Retrieve Daft DataFrame via Ray with optional column and predicate filtering\n",
    "        daft_dataframe = ray.get(vast_actor.get_vast_data.remote(columns, predicate, internal_row_id))\n",
    "        \n",
    "        print(f\"[SUCCESS] Created Daft DataFrame from {table_name}\", file=sys.stderr)\n",
    "        return daft.from_arrow(daft_dataframe)\n",
    "    \n",
    "    except Exception as e:\n",
    "        print(f\"[ERROR] Failed to create VAST table: {e}\", file=sys.stderr)\n",
    "        traceback.print_exc(file=sys.stderr)\n",
    "        raise"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "0bd8465c-beae-46d8-bfe3-4aa3dfa03183",
   "metadata": {},
   "source": [
    "## Connect to ray cluster"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 5,
   "id": "e01b1dac-4c53-4ea4-971a-1089839f5c14",
   "metadata": {},
   "outputs": [
    {
     "name": "stderr",
     "output_type": "stream",
     "text": [
      "2025-03-26 16:44:19,259\tINFO client_builder.py:244 -- Passing the following kwargs to ray.init() on the server: log_to_driver\n",
      "SIGTERM handler is not set because current thread is not the main thread.\n",
      "\u001b[36m(VastDataSource pid=65110, ip=10.1.171.2)\u001b[0m [INIT] VastDataSource initialized for table: tweets\n",
      "\u001b[36m(VastDataSource pid=65110, ip=10.1.171.2)\u001b[0m [METHOD] get_vast_data called for tweets\n",
      "\u001b[36m(VastDataSource pid=65110, ip=10.1.171.2)\u001b[0m [PARAMS] Columns: ['id', 'text'], Predicate: _.text.contains('AI')\n",
      "\u001b[36m(VastDataSource pid=65110, ip=10.1.171.2)\u001b[0m [CONNECTION] VAST database connection established\n",
      "\u001b[36m(VastDataSource pid=65110, ip=10.1.171.2)\u001b[0m [DEBUG] Starting batch reading for tweets\n",
      "\u001b[36m(VastDataSource pid=65110, ip=10.1.171.2)\u001b[0m [HOST] Hostname: raycluster-kuberay-workergroup-worker-m4vp6\n",
      "\u001b[36m(VastDataSource pid=65110, ip=10.1.171.2)\u001b[0m [DEBUG] Reader type: <class 'pyarrow.lib.RecordBatchReader'>\n",
      "\u001b[36m(VastDataSource pid=65110, ip=10.1.171.2)\u001b[0m [DEBUG] Reader schema: id: int64\n",
      "\u001b[36m(VastDataSource pid=65110, ip=10.1.171.2)\u001b[0m   -- field metadata --\n",
      "\u001b[36m(VastDataSource pid=65110, ip=10.1.171.2)\u001b[0m   VAST:column_id: '65'\n",
      "\u001b[36m(VastDataSource pid=65110, ip=10.1.171.2)\u001b[0m text: string\n",
      "\u001b[36m(VastDataSource pid=65110, ip=10.1.171.2)\u001b[0m   -- field metadata --\n",
      "\u001b[36m(VastDataSource pid=65110, ip=10.1.171.2)\u001b[0m   VAST:column_id: '67'\n",
      "\u001b[36m(VastDataSource pid=65110, ip=10.1.171.2)\u001b[0m $row_id: uint64\n",
      "\u001b[36m(VastDataSource pid=65110, ip=10.1.171.2)\u001b[0m [DEBUG] Collected batch 0\n",
      "\u001b[36m(VastDataSource pid=65110, ip=10.1.171.2)\u001b[0m [DEBUG] Collected batch 1\n",
      "\u001b[36m(VastDataSource pid=65110, ip=10.1.171.2)\u001b[0m [DEBUG] Collected batch 2\n",
      "\u001b[36m(VastDataSource pid=65110, ip=10.1.171.2)\u001b[0m [DEBUG] Collected batch 3\n",
      "\u001b[36m(VastDataSource pid=65110, ip=10.1.171.2)\u001b[0m [DEBUG] Batch reading summary:\n",
      "\u001b[36m(VastDataSource pid=65110, ip=10.1.171.2)\u001b[0m   Total batches: 4\n",
      "\u001b[36m(VastDataSource pid=65110, ip=10.1.171.2)\u001b[0m   Total time: 0.40 seconds\n",
      "\u001b[36m(VastDataSource pid=65110, ip=10.1.171.2)\u001b[0m [SUCCESS] Retrieved table with 130600 rows\n"
     ]
    }
   ],
   "source": [
    "import ray\n",
    "from daft import context\n",
    "\n",
    "# Initialize Ray with the specified address\n",
    "if not ray.is_initialized():\n",
    "    ray.init(address=\"ray://10.143.11.241:30001\", runtime_env={\"pip\": [\"daft\", \"vastdb\", \"pyarrow\"]})\n",
    "\n",
    "    # set daft to use ray runner\n",
    "    if not context.get_context():\n",
    "        context.set_runner_ray(address=\"ray://10.143.11.241:30001\")\n"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "196ce365-4684-4f9a-952c-71bf43b3b6dd",
   "metadata": {},
   "source": [
    "## Ingest Vast DB table as Daft Dataframe"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 6,
   "id": "c89dc135-6155-4451-a37c-1cb8cf6acad8",
   "metadata": {},
   "outputs": [
    {
     "name": "stderr",
     "output_type": "stream",
     "text": [
      "[FUNCTION] create_vast_table called for tweets\n",
      "[SUCCESS] Created Daft DataFrame from tweets\n"
     ]
    }
   ],
   "source": [
    "from ibis import _\n",
    "\n",
    "# Create Daft DataFrame from VAST\n",
    "df = fetch_as_daft_dataframe (\n",
    "    endpoint='http://172.200.204.2:80',\n",
    "    access_key='Y5101AQQTB1PUAEKQXN5',\n",
    "    secret_key='bsqwYOcsvfXxsvtTYruCT24c3w1E1Y8iBpmyoLGr',\n",
    "    bucket_name='csnow-db',\n",
    "    schema_name='social_media',\n",
    "    table_name='tweets',\n",
    "    columns=['id', 'text'],\n",
    "    predicate=(_.text.contains('AI')),\n",
    "    internal_row_id=True\n",
    ")"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "53dde5c6-845a-40f9-bdda-a476f637e090",
   "metadata": {},
   "source": [
    "## Basic Data Engineering on the Daft Dataframe\n",
    "\n",
    "We aren't really doing anything substantial here - see the Daft [DataFrame API](https://www.getdaft.io/projects/docs/en/stable/api_docs/dataframe.html) for more interesing methods."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 7,
   "id": "25af3fc8-94a8-4ef8-88ae-a5246c8a2f4f",
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "<div>\n",
       "<table class=\"dataframe\">\n",
       "<thead><tr><th style=\"text-wrap: nowrap; max-width:192px; overflow:auto; text-align:left\">id<br />Int64</th><th style=\"text-wrap: nowrap; max-width:192px; overflow:auto; text-align:left\">text<br />Utf8</th><th style=\"text-wrap: nowrap; max-width:192px; overflow:auto; text-align:left\">$row_id<br />UInt64</th></tr></thead>\n",
       "<tbody>\n",
       "<tr><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">4972466409915944915</div></td><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">can't believe how incredible AI is!</div></td><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">1572879</div></td></tr>\n",
       "<tr><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">-1590415462612194030</div></td><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">impressed with how cool AI is!</div></td><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">1572903</div></td></tr>\n",
       "<tr><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">8486930471810695546</div></td><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">finally got how kind AI is!</div></td><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">1572932</div></td></tr>\n",
       "<tr><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">-8343872824439823953</div></td><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">ready to see how wonderful AI is!</div></td><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">1573017</div></td></tr>\n",
       "<tr><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">5043897367125593095</div></td><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">motivated by how awesome AI is!</div></td><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">1573018</div></td></tr>\n",
       "<tr><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">-1657734256534097438</div></td><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">just discovered how powerful AI is!</div></td><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">1573030</div></td></tr>\n",
       "<tr><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">-1815539176794851507</div></td><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">can't wait to see how brilliant AI is!</div></td><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">1573053</div></td></tr>\n",
       "<tr><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">-5287482691421413087</div></td><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">can't believe how wonderful AI is!</div></td><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">1573131</div></td></tr>\n",
       "</tbody>\n",
       "</table>\n",
       "<small>(Showing first 8 of 130600 rows)</small>\n",
       "</div>"
      ],
      "text/plain": [
       "╭──────────────────────┬────────────────────────────────┬─────────╮\n",
       "│ id                   ┆ text                           ┆ $row_id │\n",
       "│ ---                  ┆ ---                            ┆ ---     │\n",
       "│ Int64                ┆ Utf8                           ┆ UInt64  │\n",
       "╞══════════════════════╪════════════════════════════════╪═════════╡\n",
       "│ 4972466409915944915  ┆ can't believe how incredible … ┆ 1572879 │\n",
       "├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┤\n",
       "│ -1590415462612194030 ┆ impressed with how cool AI is… ┆ 1572903 │\n",
       "├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┤\n",
       "│ 8486930471810695546  ┆ finally got how kind AI is!    ┆ 1572932 │\n",
       "├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┤\n",
       "│ -8343872824439823953 ┆ ready to see how wonderful AI… ┆ 1573017 │\n",
       "├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┤\n",
       "│ 5043897367125593095  ┆ motivated by how awesome AI i… ┆ 1573018 │\n",
       "├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┤\n",
       "│ -1657734256534097438 ┆ just discovered how powerful … ┆ 1573030 │\n",
       "├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┤\n",
       "│ -1815539176794851507 ┆ can't wait to see how brillia… ┆ 1573053 │\n",
       "├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┤\n",
       "│ -5287482691421413087 ┆ can't believe how wonderful A… ┆ 1573131 │\n",
       "╰──────────────────────┴────────────────────────────────┴─────────╯\n",
       "\n",
       "(Showing first 8 of 130600 rows)"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "df.show()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 8,
   "id": "f33ac421-44ee-4a85-a20e-f34fd1f5c4e5",
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "<div>\n",
       "<table class=\"dataframe\">\n",
       "<thead><tr><th style=\"text-wrap: nowrap; max-width:192px; overflow:auto; text-align:left\">count<br />UInt64</th></tr></thead>\n",
       "<tbody>\n",
       "<tr><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">130600</div></td></tr>\n",
       "</tbody>\n",
       "</table>\n",
       "<small>(Showing first 1 of 1 rows)</small>\n",
       "</div>"
      ],
      "text/plain": [
       "╭────────╮\n",
       "│ count  │\n",
       "│ ---    │\n",
       "│ UInt64 │\n",
       "╞════════╡\n",
       "│ 130600 │\n",
       "╰────────╯\n",
       "\n",
       "(Showing first 1 of 1 rows)"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "df.count().show()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 9,
   "id": "8cea00d2-218e-4447-b291-9f4f26caf8df",
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "<div>\n",
       "<table class=\"dataframe\">\n",
       "<thead><tr><th style=\"text-wrap: nowrap; max-width:192px; overflow:auto; text-align:left\">id<br />Int64</th><th style=\"text-wrap: nowrap; max-width:192px; overflow:auto; text-align:left\">text<br />Utf8</th><th style=\"text-wrap: nowrap; max-width:192px; overflow:auto; text-align:left\">$row_id<br />UInt64</th></tr></thead>\n",
       "<tbody>\n",
       "<tr><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">4972466409915944915</div></td><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">can't believe how incredible AI is!</div></td><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">1572879</div></td></tr>\n",
       "<tr><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">-1590415462612194030</div></td><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">impressed with how cool AI is!</div></td><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">1572903</div></td></tr>\n",
       "<tr><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">8486930471810695546</div></td><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">finally got how kind AI is!</div></td><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">1572932</div></td></tr>\n",
       "<tr><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">-8343872824439823953</div></td><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">ready to see how wonderful AI is!</div></td><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">1573017</div></td></tr>\n",
       "<tr><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">5043897367125593095</div></td><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">motivated by how awesome AI is!</div></td><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">1573018</div></td></tr>\n",
       "</tbody>\n",
       "</table>\n",
       "<small>(Showing first 5 of 5 rows)</small>\n",
       "</div>"
      ],
      "text/plain": [
       "╭──────────────────────┬────────────────────────────────┬─────────╮\n",
       "│ id                   ┆ text                           ┆ $row_id │\n",
       "│ ---                  ┆ ---                            ┆ ---     │\n",
       "│ Int64                ┆ Utf8                           ┆ UInt64  │\n",
       "╞══════════════════════╪════════════════════════════════╪═════════╡\n",
       "│ 4972466409915944915  ┆ can't believe how incredible … ┆ 1572879 │\n",
       "├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┤\n",
       "│ -1590415462612194030 ┆ impressed with how cool AI is… ┆ 1572903 │\n",
       "├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┤\n",
       "│ 8486930471810695546  ┆ finally got how kind AI is!    ┆ 1572932 │\n",
       "├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┤\n",
       "│ -8343872824439823953 ┆ ready to see how wonderful AI… ┆ 1573017 │\n",
       "├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┤\n",
       "│ 5043897367125593095  ┆ motivated by how awesome AI i… ┆ 1573018 │\n",
       "╰──────────────────────┴────────────────────────────────┴─────────╯\n",
       "\n",
       "(Showing first 5 of 5 rows)"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "df.limit(5).show()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 10,
   "id": "40e359d2-dd34-4aa0-a0d1-e97053d51408",
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "<div>\n",
       "<table class=\"dataframe\">\n",
       "<thead><tr><th style=\"text-wrap: nowrap; max-width:192px; overflow:auto; text-align:left\">id<br />Int64</th><th style=\"text-wrap: nowrap; max-width:192px; overflow:auto; text-align:left\">text<br />Utf8</th><th style=\"text-wrap: nowrap; max-width:192px; overflow:auto; text-align:left\">$row_id<br />UInt64</th></tr></thead>\n",
       "<tbody>\n",
       "<tr><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">-1002806103367248550</div></td><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">totally in love with how perfect AI is!</div></td><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">1573338</div></td></tr>\n",
       "<tr><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">4483608969875055006</div></td><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">totally in love with how clever AI is!</div></td><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">1573501</div></td></tr>\n",
       "<tr><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">-3485059169440733923</div></td><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">looking forward to see how lovely AI is!</div></td><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">1573670</div></td></tr>\n",
       "<tr><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">9061853070521936991</div></td><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">totally in love with how perfect AI is!</div></td><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">1573721</div></td></tr>\n",
       "<tr><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">4765031840801089800</div></td><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">just discovered how lovely AI is!</div></td><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">1573810</div></td></tr>\n",
       "<tr><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">-195608600448131006</div></td><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">totally in love with how lovely AI is!</div></td><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">1574541</div></td></tr>\n",
       "<tr><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">-3990254773133657667</div></td><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">totally in love with how awesome AI is!</div></td><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">1574564</div></td></tr>\n",
       "<tr><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">-7108812032965286885</div></td><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">totally in love with how beautiful AI is!</div></td><td><div style=\"text-align:left; max-width:192px; max-height:64px; overflow:auto\">1574910</div></td></tr>\n",
       "</tbody>\n",
       "</table>\n",
       "<small>(Showing first 8 rows)</small>\n",
       "</div>"
      ],
      "text/plain": [
       "╭──────────────────────┬────────────────────────────────┬─────────╮\n",
       "│ id                   ┆ text                           ┆ $row_id │\n",
       "│ ---                  ┆ ---                            ┆ ---     │\n",
       "│ Int64                ┆ Utf8                           ┆ UInt64  │\n",
       "╞══════════════════════╪════════════════════════════════╪═════════╡\n",
       "│ -1002806103367248550 ┆ totally in love with how perf… ┆ 1573338 │\n",
       "├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┤\n",
       "│ 4483608969875055006  ┆ totally in love with how clev… ┆ 1573501 │\n",
       "├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┤\n",
       "│ -3485059169440733923 ┆ looking forward to see how lo… ┆ 1573670 │\n",
       "├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┤\n",
       "│ 9061853070521936991  ┆ totally in love with how perf… ┆ 1573721 │\n",
       "├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┤\n",
       "│ 4765031840801089800  ┆ just discovered how lovely AI… ┆ 1573810 │\n",
       "├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┤\n",
       "│ -195608600448131006  ┆ totally in love with how love… ┆ 1574541 │\n",
       "├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┤\n",
       "│ -3990254773133657667 ┆ totally in love with how awes… ┆ 1574564 │\n",
       "├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┤\n",
       "│ -7108812032965286885 ┆ totally in love with how beau… ┆ 1574910 │\n",
       "╰──────────────────────┴────────────────────────────────┴─────────╯\n",
       "\n",
       "(Showing first 8 rows)"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "daft.sql(\"SELECT * FROM df WHERE text like '%love%'\").show()"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "37186c61-00af-4fbb-8230-176b0e635a54",
   "metadata": {},
   "source": [
    "## Train ML model"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "bf29ac73-2aee-430c-a896-eeb85803c6a1",
   "metadata": {},
   "source": [
    "Okay, the model is a bit useless, but you get to see the mechanics of using Ray Train on Vast DB data."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 11,
   "id": "d9118e19-e232-4d35-98e8-7cdbff227788",
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "ray.data.dataset.MaterializedDataset"
      ]
     },
     "execution_count": 11,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "# Convert the Daft dataframe to a Ray Dataset\n",
    "\n",
    "rdf = df.to_ray_dataset()\n",
    "type(rdf)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "c248e5fe-b39d-488a-9021-ad6ffafcf8d8",
   "metadata": {},
   "source": [
    "Coming soon Ray Train [example](https://docs.ray.io/en/latest/train/train.html) ..."
   ]
  },
  {
   "cell_type": "markdown",
   "id": "0f280c43-1843-46f0-9bab-06dc2193e08b",
   "metadata": {},
   "source": [
    "## Shutdown Ray client\n",
    "\n",
    "Shutdown the ray client when you have finished using it.\n",
    "\n",
    "You can also achieve the same by stopping the jupyter notebook skernel."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 12,
   "id": "24349c42-5386-4b30-9304-4dec90b66119",
   "metadata": {},
   "outputs": [],
   "source": [
    "# ray.shutdown()"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3 (ipykernel)",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.9.21"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}
